from datetime import timedelta

from django.conf import settings

import environ


class Env(environ.Env):

    def timedelta(self, var, default=environ.Env.NOTSET):

        def cast(value):
            args = [val.split('=') for val in value.split(',') if val]
            return timedelta(**{name: float(val) for name, val in args})

        return self.get_value(var, cast=cast, default=default)


env = Env()

JWT_ALGORITHM = env('JWT_ALGORITHM', default='HS256')
JWT_AUDIENCE = env('JWT_AUDIENCE', default=None)
JWT_AUTH_HEADER_PREFIX = env('JWT_AUTH_HEADER_PREFIX', default='JWT')
JWT_ISSUER = env('JWT_ISSUER', default=None)
JWT_LEEWAY = env.timedelta('JWT_LEEWAY', 0)
JWT_SECRET_KEY = env('JWT_SECRET_KEY', default=settings.SECRET_KEY)

JWT_VERIFY = env.bool('JWT_VERIFY', True)
JWT_VERIFY_EXPIRATION = env.bool('JWT_VERIFY_EXPIRATION', False)

JWT_EXPIRATION_DELTA = env.timedelta(
    'JWT_EXPIRATION_DELTA',
    timedelta(seconds=60 * 5),
)

JWT_ALLOW_REFRESH = env.bool('JWT_ALLOW_REFRESH', True)
JWT_REFRESH_EXPIRATION_DELTA = env.timedelta(
    'JWT_REFRESH_EXPIRATION_DELTA',
    timedelta(days=7),
)



from datetime import timedelta
from unittest.mock import Mock, PropertyMock, patch

from django.contrib.auth import get_user_model
from django.test import RequestFactory, TestCase

from graphql_jwt import utils
from graphql_jwt.exceptions import GraphQLJWTError

from .decorators import override_settings


class UtilsTests(TestCase):

    def setUp(self):
        self.user = get_user_model().objects.create_user(username='test')
        self.factory = RequestFactory()

    @patch('django.contrib.auth.models.User.get_username',
           return_value=Mock(pk='test'))
    def test_payload_foreign_key_pk(self, *args):
        payload = utils.jwt_payload(self.user)
        username_field = get_user_model().USERNAME_FIELD

        self.assertEqual(payload[username_field], 'test')

    @override_settings(JWT_AUDIENCE='test')
    def test_payload_audience(self):
        payload = utils.jwt_payload(self.user)
        self.assertEqual(payload['aud'], 'test')

    @override_settings(JWT_ISSUER='test')
    def test_payload_issuer(self):
        payload = utils.jwt_payload(self.user)
        self.assertEqual(payload['iss'], 'test')

    def test_invalid_authorization_header_prefix(self):
        headers = {
            'HTTP_AUTHORIZATION': 'INVALID token',
        }

        request = self.factory.get('/', **headers)
        header = utils.get_authorization_header(request)

        self.assertIsNone(header)

    @override_settings(JWT_VERIFY_EXPIRATION=True)
    @override_settings(JWT_EXPIRATION_DELTA=timedelta(seconds=-1))
    def test_payload_expired_signature(self):
        payload = utils.jwt_payload(self.user)
        token = utils.jwt_encode(payload)

        with self.assertRaises(GraphQLJWTError):
            utils.get_payload(token)

    def test_payload_decode_audience_missing(self):
        payload = utils.jwt_payload(self.user)
        token = utils.jwt_encode(payload)

        with override_settings(JWT_AUDIENCE='test'):
            with self.assertRaises(GraphQLJWTError):
                utils.get_payload(token)

    def test_payload_decode_error(self):
        with self.assertRaises(GraphQLJWTError):
            utils.get_payload('invalid')

    def test_user_by_natural_key_not_exists(self):
        user = utils.get_user_by_natural_key(0)
        self.assertIsNone(user)

    def test_user_by_invalid_payload(self):
        with self.assertRaises(GraphQLJWTError):
            utils.get_user_by_payload({})

    @patch('django.contrib.auth.models.User.is_active',
           new_callable=PropertyMock,
           return_value=False)
    def test_user_disabled_by_payload(self, *args):
        payload = utils.jwt_payload(self.user)

        with self.assertRaises(GraphQLJWTError):
            utils.get_user_by_payload(payload)



from datetime import timedelta
from unittest.mock import patch

from django.test import TestCase

from graphql_jwt import settings


class SettingsTests(TestCase):

    @patch.dict('os.environ', {
        'TEST_TIMEDELTA': 'hours=1,days=1',
    })
    def test_env_timedelta_cast(self):
        env = settings.Env()
        var = env.timedelta('TEST_TIMEDELTA')

        self.assertEqual(var, timedelta(hours=1, days=1))



from graphql_jwt import settings as graphql_jwt_settings
from graphql_jwt.backends import JWTBackend
from graphql_jwt.exceptions import GraphQLJWTError

from .base import GraphQLJWTTestCase


class BackendsTests(GraphQLJWTTestCase):

    def test_authenticate(self):
        headers = {
            'HTTP_AUTHORIZATION': '{0} {1}'.format(
                graphql_jwt_settings.JWT_AUTH_HEADER_PREFIX,
                self.token),
        }

        request = self.factory.get('/', **headers)
        user = JWTBackend().authenticate(request=request)

        self.assertEqual(user, self.user)

    def test_authenticate_fail(self):
        headers = {
            'HTTP_AUTHORIZATION': '{} invalid'.format(
                graphql_jwt_settings.JWT_AUTH_HEADER_PREFIX),
        }

        request = self.factory.get('/', **headers)

        with self.assertRaises(GraphQLJWTError):
            JWTBackend().authenticate(request=request)

    def test_authenticate_null_request(self):
        user = JWTBackend().authenticate(request=None)
        self.assertIsNone(user)

    def test_authenticate_missing_token(self):
        request = self.factory.get('/')
        user = JWTBackend().authenticate(request=request)

        self.assertIsNone(user)

    def test_get_user(self):
        user = JWTBackend().get_user(self.user.username)
        self.assertEqual(user, self.user)



import contextlib

from graphql_jwt import settings


@contextlib.contextmanager
def override_settings(**override):
    restore = {}

    try:
        for var, value in override.items():
            restore[var] = getattr(settings, var)
            setattr(settings, var, value)
        yield
    finally:
        for var, value in restore.items():
            setattr(settings, var, value)



from datetime import datetime, timedelta
from unittest.mock import patch

from graphql_jwt import settings
from graphql_jwt.utils import jwt_encode, jwt_payload

from .base import GraphQLJWTTestCase
from .decorators import override_settings


class MutationsTests(GraphQLJWTTestCase):

    def test_verify(self):
        query = '''
        mutation VerifyToken($token: String!) {
          verifyToken(token: $token) {
            payload
          }
        }'''

        response = self.client.execute(query, token=self.token)
        payload = response.data['verifyToken']['payload']

        self.assertEqual(self.user.username, payload['username'])

    def test_verify_invalid_token(self):
        query = '''
        mutation VerifyToken($token: String!) {
          verifyToken(token: $token) {
            payload
          }
        }'''

        response = self.client.execute(query, token='invalid')
        self.assertTrue(response.errors)

    def test_refresh(self):
        query = '''
        mutation RefreshToken($token: String!) {
          refreshToken(token: $token) {
            data
          }
        }'''

        with patch('graphql_jwt.utils.datetime') as datetime_mock:
            datetime_mock.utcnow.return_value =\
                datetime.utcnow() + timedelta(seconds=1)

            response = self.client.execute(query, token=self.token)

        data = response.data['refreshToken']['data']

        self.assertNotEqual(self.token, data['token'])
        self.assertEqual(self.user.username, data['payload']['username'])

    def test_refresh_expired(self):
        query = '''
        mutation RefreshToken($token: String!) {
          refreshToken(token: $token) {
            data
          }
        }'''

        with patch('graphql_jwt.mutations.datetime') as datetime_mock:
            datetime_mock.utcnow.return_value = datetime.utcnow() +\
                settings.JWT_REFRESH_EXPIRATION_DELTA +\
                timedelta(seconds=1)

            response = self.client.execute(query, token=self.token)

        self.assertTrue(response.errors)

    @override_settings(JWT_ALLOW_REFRESH=False)
    def test_refresh_error(self, *args):
        payload = jwt_payload(self.user)
        token = jwt_encode(payload)

        query = '''
        mutation RefreshToken($token: String!) {
          refreshToken(token: $token) {
            data
          }
        }'''

        response = self.client.execute(query, token=token)
        self.assertTrue(response.errors)
